use std::io;
use std::time::{Duration, Instant};

use crossterm::event;
use crossterm::event::{Event, KeyCode};
use tui::backend::Backend;
use tui::layout::*;
use tui::text::*;
use tui::widgets::*;
use tui::{Frame, Terminal};

use crate::consts::*;
use crate::ui_screens::utils::centered_rect_size;
use crate::ui_screens::ScreenResult;
use crate::ui_widgets::button_widget::ButtonWidget;

/// Convenience function to inform user that his terminal window is too small to display the current
/// screen
pub fn show_screen_too_small_popup<B: Backend>(f: &mut Frame<B>) {
    PopupScreen::new("🖵  Screen too small!").show_in_frame(f)
}

pub struct PopupScreen<'a> {
    title: &'a str,
    msg: &'a str,
    can_close: bool,
}

impl<'a> PopupScreen<'a> {
    pub fn new(msg: &'a str) -> Self {
        Self {
            title: "Message",
            msg,
            can_close: true,
        }
    }

    pub fn set_can_close(mut self, can_close: bool) -> Self {
        self.can_close = can_close;
        self
    }

    pub fn show<B: Backend>(mut self, terminal: &mut Terminal<B>) -> io::Result<ScreenResult<()>> {
        let mut last_tick = Instant::now();
        loop {
            terminal.draw(|f| self.ui(f))?;

            let timeout = TICK_RATE
                .checked_sub(last_tick.elapsed())
                .unwrap_or_else(|| Duration::from_secs(0));

            if event::poll(timeout)? {
                if let Event::Key(key) = event::read()? {
                    match key.code {
                        KeyCode::Char('q') | KeyCode::Esc => return Ok(ScreenResult::Canceled),
                        KeyCode::Enter => {
                            return Ok(ScreenResult::Ok(()));
                        }
                        _ => {}
                    }
                }
            }
            if last_tick.elapsed() >= TICK_RATE {
                last_tick = Instant::now();
            }
        }
    }

    /// Show message once message, without polling messages
    pub fn show_once<B: Backend>(mut self, terminal: &mut Terminal<B>) -> io::Result<()> {
        self.can_close = false;
        terminal.draw(|f| self.ui(f))?;
        Ok(())
    }

    /// Show message once message in a frame, without polling messages
    pub fn show_in_frame<B: Backend>(mut self, frame: &mut Frame<B>) {
        self.can_close = false;
        self.ui(frame)
    }

    fn ui<B: Backend>(&mut self, f: &mut Frame<B>) {
        // Preprocess message
        let lines = textwrap::wrap(self.msg, f.size().width as usize - 20);
        let line_max_len = lines.iter().map(|l| l.len()).max().unwrap();

        let area = centered_rect_size(
            line_max_len as u16 + 4,
            match self.can_close {
                // reserve space for button
                true => 5,
                false => 2,
            } + lines.len() as u16,
            &f.size(),
        );

        let block = Block::default().borders(Borders::ALL).title(self.title);
        f.render_widget(block, area);

        // Create two chunks with equal horizontal screen space
        let chunks = Layout::default()
            .direction(Direction::Vertical)
            .constraints(
                [
                    Constraint::Length(lines.len() as u16),
                    Constraint::Length(3),
                ]
                .as_ref(),
            )
            .split(area.inner(&Margin {
                horizontal: 2,
                vertical: 1,
            }));

        let text = lines
            .iter()
            .map(|s| Spans::from(s.as_ref()))
            .collect::<Vec<_>>();
        let paragraph = Paragraph::new(text);
        f.render_widget(paragraph, chunks[0]);

        if self.can_close {
            let ok_button = ButtonWidget::new("OK", true);
            f.render_widget(ok_button, chunks[1]);
        }
    }
}
